###########################################################################
# <LICENSEINFO>
# Copyright (C) Daniel Guermeur - 1999-2001
# Copyright (C) Metadot Corporation - 2001-2004
# 
# This file is part of Metadot Portal Server software $Name: Release_6_1_6 $
# 
# Metadot Portal Server software is free software; you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
# 
# Metadot Portal Server software is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# Metadot Portal Server; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# For more information, please contact info@metadot.com.
#</LICENSEINFO>
############################################################################

# The DistributedCache shares update information across server instances.
# This cache manages dirtiness at the element level
                    
package Metadot::ObjectAccess::DistributedCache;
use strict;
use vars qw(@ISA);

use Metadot qw(%FORM $SESSION $PARAMS %REQUEST_VARS);
use MetadotObject;
use Metadot::SystemMessage;
use Metadot::SystemMessageManager;
use Data::Dumper;
use Metadotclass;

@ISA = qw(MetadotObject);

sub new {
    my $proto = shift;
    my %args = @_;

    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(%args);
    
    bless $self, $class;
    
    return $self;
}


sub _initialize {
    my $self = shift;
    my %args = @_;
    $self->SUPER::_initialize(%args);
    
    $self->{_element_hashtable} = {};
    $self->{_first_node}        = undef;
    $self->{_last_node}         = undef;
    $self->{_size}              = 0;

    $self->{creation_time} = time;
    $self->{_hits}      = 0;
    $self->{_misses}    = 0;
}


sub get_constructor_properties {
    return {
        owner           => { value=> undef},    # the class that owns the cache object
        signature       => { value=> undef},    # the signature of the cache, used to identify
                                                #     like instances of cache
        max_size        => { value=> undef},    # the maximum number of elements in the cache
        max_element_age => { value=> 3600},     # maximum age before an element is flushed from cache
    };
}

###############################################################

# puts an element into the cache
sub put {
    my ($self, $key, $value) = @_;
    defined($key) || die("Must specify the key for placing an element in the cache");

    # before I process the cache, see if I have any new notifications
    $self->initialize_request();
    
    # check if this key is in use
    my $hashtable = $self->_get_element_hashtable();
    my $existing  = $hashtable->{$key};
    
    # if the key is in use, update the element
    if ($existing) {
        Metadotclass->debug_msg(4, "DistributedCache::put() Updating existing value, key='$key', value=$value");
        $self->_debug_element($existing);
        $existing->{value} = $value;
        $self->_set_MRU($existing);
    } else {
        my $element = { value=>$value, timestamp=>time, key=>$key };
        Metadotclass->debug_msg(4, "DistributedCache::put() Adding new value, key='$key', value=$value");
        $self->_debug_element($element);
        $hashtable->{$key} = $element;
        $self->_add_to_list($element);
    }
}

# retrieves an element from the cache
sub get {
    my ($self, $key) = @_;
    defined($key) || die("Must specify the key for placing an element in the cache");
    my $result;

    # before I process the cache, see if I have any new notifications
    $self->initialize_request();

    my $hashtable = $self->_get_element_hashtable();
    my $element = $hashtable->{$key};
    if ($element) {
        # first check that element hasn't expired
        my $age = time - $element->{timestamp};
        my $max_element_age = $self->get_max_element_age() || 0;
        if ($max_element_age && ($age > $max_element_age)) {
            Metadotclass->debug_msg(4, "DistributedCache::get() Key '$key' has expired");
            $self->delete($key);
            $self->_set_misses($self->misses() + 1);
        } else {
            $self->_set_MRU($element);
            $result = $element->{value};
            $self->_set_hits($self->hits() + 1);
            Metadotclass->debug_msg(4, "DistributedCache::get() Successfully retrieved element, key='$key' ($element)");
        }        
    } else {
        $self->_set_misses($self->misses() + 1);
        Metadotclass->debug_msg(4, "DistributedCache::get() Cache miss on key '$key'");
    }
    return $result;
}

# deletes an element from the cache
sub delete {
    my ($self, $key) = @_;
    defined($key) || die("Must specify the key for deleting an element in the cache");
    
    # before I process the cache, see if I have any new notifications
    $self->initialize_request();
    
    Metadotclass->debug_msg(4, "DistributedCache::delete() Deleting an element with key '$key'");
    my $hashtable = $self->_get_element_hashtable();
    my $element = $hashtable->{$key};

    if ($element) {
        $self->_delete_from_list($element);
        delete $hashtable->{$key};
    }
}


# measures and returns the number of elements in the cache
sub size {
    my $self = shift;
    
    # before I process the cache, see if I have any new notifications
    $self->initialize_request();
    
    return $self->access_property('_size', 0);
}

# returns the number of cache hits
sub hits {
    my $self = shift;
    
    # before I process the cache, see if I have any new notifications
    $self->initialize_request();
    
    return $self->access_property('_hits', 0);
}

# returns the number of cache misses
sub misses {
    my $self = shift;
    
    # before I process the cache, see if I have any new notifications
    $self->initialize_request();
    
    return $self->access_property('_misses', 0);
}


# completely empties the cache
sub flush {
    my $self = shift;
    
    # before I process the cache, see if I have any new notifications
    $self->initialize_request();
    
    $self->_set_size(0);
    $self->_set_element_hashtable({});
    $self->_set_first_node(undef);
    $self->_set_last_node(undef);
    $self->_set_hits(0);
    $self->_set_misses(0);

    Metadotclass->debug_msg(3, "DistributedCache::flush() Cache was flushed");
}

sub check_max_size {
    my $self = shift;
    # if the maximum size is reached, delete something first
    my $max_size = $self->get_max_size() || 0;
    while ($max_size && ($self->size() > $max_size)) {
        my $lru = $self->_get_LRU();
        $lru || die("Trying to trim cache, but LRU could not be found");
        Metadotclass->debug_msg(4, "DistributedCache::_add_to_list() Maximum size of $max_size has been exceeded. Trimming LRU, key='$lru->{element}->{key}'");
        $self->delete($lru->{key});
    }
}


# some actions should be performed prior to every request
sub initialize_request {
    my $self = shift;
    # this is a very hacky way to do request-scoped variables, but it works
    my $request_key = "__$self"."__request";
    unless ($REQUEST_VARS{$request_key}) {
        $REQUEST_VARS{$request_key} = 1;
        $self->check_notifications();
        $self->check_max_size();
    }
}

sub check_notifications {
    my $self = shift;
    # note, I calculate the time since last update instead of directly
    # sending any time values to the message manager. This is because
    # the two may be on systems with differing times.
    my $last_update_time = $self->get_last_update_time();
    $self->set_last_update_time(time-1);  # roll back by one second, for safety
    #Metadotclass->debug_msg(4, "DistributedCache::check_notifications; last_update_time was $last_update_time; current time=".time);
    Metadotclass->debug_msg(4, "DistributedCache::check_notifications; ($self)<br/>\nlast update: ".localtime($last_update_time)."<br/>\ncurrent time: ".localtime(time));
    
    if (defined($last_update_time)) {
        my $time_since_last_update = time - $last_update_time;  
        Metadotclass->debug_msg(4, "DistributedCache::check_notifications within last $time_since_last_update seconds");

        # check for updates
        my @messages = Metadot::SystemMessageManager->check_messages(
            sender  => $self->get_owner()->get_signature(),
            max_age => ($time_since_last_update),
            );
        
        foreach my $message (@messages) {
            my $content = $message->get_content();
            Metadotclass->debug_msg(4, "DistributedCache::check_notifications() Got message $message, content='$content'");
            my ($command, @args) = split(/\s+/, $content);
            $command = uc($command);
            Metadotclass->debug_msg(4, "DistributedCache::check_notifications() Parsed command='$command' args=@args");
            if ($command) {
                if ($command eq 'FLUSH') {
                    $self->flush();
                } elsif ($command eq 'DROP') {
                    # default command is a key value to delete
                    my $key = $args[0];
                    Metadotclass->debug_msg(4, "DistributedCache::check_notifications() Somebody updated element '$key'");
                    $self->delete($key);
                }
            }
        }
    }    
}


###############################################################
# public accessors / mutators

sub set_creation_time {
    my $self = shift;
    $self->mutate_property('creation_time', @_);
}

sub get_creation_time {
    my $self = shift;
    return $self->access_property('creation_time', @_);
}

sub set_max_element_age {
    my $self = shift;
    $self->mutate_property('max_element_age', @_);
}

sub get_max_element_age {
    my $self = shift;
    return $self->access_property('max_element_age', $PARAMS->{accelerator_max_element_age});
}

sub set_last_update_time {
    my $self = shift;
    $self->mutate_property('last_update_time', @_);
}

sub get_last_update_time {
    my $self = shift;
    return $self->access_property('last_update_time', @_);
}

sub set_owner {
    my $self = shift;
    $self->mutate_property('owner', @_);
}

sub get_owner {
    my $self = shift;
    return $self->access_property('owner', @_);
}

sub set_signature {
    my $self = shift;
    $self->mutate_property('signature', @_);
}

sub get_signature {
    my $self = shift;
    return $self->access_property('signature', @_);
}

sub set_max_size {
    my $self = shift;
    $self->mutate_property('max_size', @_);
}

sub get_max_size {
    my $self = shift;
    return $self->access_property('max_size', $PARAMS->{accelerator_max_size});
}


###############################################################
# private operations

# adds an element to the element list
sub _add_to_list {
    my ($self, $element) = @_;
    defined($element) || die("Element not defined");
    
    # each element has a list node, and these are doubly linked.
    my $list_node = $element->{list_node};
    # if the element already has a list node, this is taken care of
    unless ($list_node) {
        $list_node = { element=> $element };
        $element->{list_node} = $list_node;
    }

    # get the first element in the list
    my $first = $self->_get_first_node();
    
    # if one was found, give it a link to the new first element
    if ($first) {
        $first->{previous_node} = $list_node;
        $list_node->{next_node} = $first;
    }

    # designate the new element as the first element
    $self->_set_first_node($list_node);

    # if there was no last element, then set this one as last
    unless ($self->_get_last_node()) {
        $self->_set_last_node($list_node);
    }

    # adjust the size
    $self->_set_size($self->size() + 1);    
}

# deletes an element from the element list
sub _delete_from_list {
    my ($self, $element) = @_;
    defined($element) || die("Element not defined");
    
    my $list_node = $element->{list_node} || die("Could not resolve list node from element");

    # get the previous and next
    my $previous = $list_node->{previous_node};
    my $next     = $list_node->{next_node};

    # unlink this node
    $list_node->{previous_node} = undef;
    $list_node->{next_node} = undef;

    # stitch the links between previous and next
    if ($previous) {
        # link the previous to the next
        if ($next) {
            $previous->{next_node} = $next;
        # unless there is not one, in which case this is now the last one
        } else {
            $self->_set_last_node($previous);
        }
    } else {
        # this is the first node. reset the list's first_node
        $self->_set_first_node($next);
    }

    if ($next) {
        # link the next to the previous
        if ($previous) {
            $next->{previous_node} = $previous;
        # unless there is no previous, in which case this is now the first one
        } else {
            $self->_set_first_node($next);
        }
    } else {
        # this is the last node. reset the list's last_node
        $self->_set_last_node($previous);
    }
    
    # adjust the size
    $self->_set_size( $self->size() - 1);
    $self->_show_list_info();
}


# sets the MRU element
sub _set_MRU {
    my $self = shift;
    my $element = shift || die("Specify an element to set_MRU");
    Metadotclass->debug_msg(5, "DistributedCache::_set_MRU key='$element->{key}'");
    $self->_delete_from_list($element);
    $self->_add_to_list($element);
    $self->_show_list_info();
}


# gets the LRU element
sub _get_LRU {
    my $self = shift;
    return $self->_get_last_node()->{element};
}


# provides an abbreviated dump of an element
sub _debug_element {
    my $self = shift;
    
    return unless ($SESSION->{debug_level} >= 5);
    
    my $element = shift;
    if ($element) {
        if (ref($element) eq 'HASH') {
            Metadotclass->debug_msg(5, "DistributedCache Dumping element $element");
            foreach my $key (keys %$element) {
                Metadotclass->debug_msg(5, "$key = $element->{$key}");
            }
            Metadotclass->debug_msg(5, ('-'x20));
        } else {
            Metadotclass->debug_msg(5, "Element is not a hash");
        }
    } else {
        Metadotclass->debug_msg(5, "Not an element");
    }
}


sub _show_list_info {
    my $self = shift;
    
    return unless ($SESSION->{debug_level} >= 6);

    my $first = $self->_get_first_node();
    my $last = $self->_get_last_node();
    my $size = $self->size();
    
    Metadotclass->debug_msg(6, "DistributedCache showing list info:");
    Metadotclass->debug_msg(6, "First: $first->{element}->{key}");
    Metadotclass->debug_msg(6, "Last : $last->{element}->{key}");
    Metadotclass->debug_msg(6, "Size : $size");
    Metadotclass->debug_msg(6, ('-'x20));
    my $node = $first;
    my $count = 1;
    while ($node) {
        Metadotclass->debug_msg(6, "List index #$count: $node->{element}->{key}");
        $node = $node->{next_node};
    }
    Metadotclass->debug_msg(6, ('-'x20));
}


###############################################################
# private accessors / mutators

sub _get_element_hashtable {
    my $self = shift;
    return $self->access_property('_element_hashtable', @_);
}

sub _set_element_hashtable {
    my $self = shift;
    return $self->mutate_property('_element_hashtable', @_);
}

# gets the first element in the list
sub _get_first_node {
    my $self = shift;
    return $self->access_property('_first_node', @_);
}


# sets the first element in the list
sub _set_first_node {
    my $self = shift;
    my $node = shift;
    if (defined($node)) {
        $node->{previous_node} = undef;
    }
    return $self->mutate_property('_first_node', $node, @_);
}


# gets the last element in the list
sub _get_last_node {
    my $self = shift;
    return $self->access_property('_last_node', @_);
}


# sets the last element in the list
sub _set_last_node {
    my $self = shift;
    my $node = shift;
    if (defined($node)) {
        $node->{next_node} = undef;
    }
    return $self->mutate_property('_last_node', $node, @_);
}

sub _set_size {
    my $self = shift;
    $self->mutate_property('_size', @_);
}

sub _set_misses {
    my $self = shift;
    $self->mutate_property('_misses', @_);
}

sub _set_hits {
    my $self = shift;
    $self->mutate_property('_hits', @_);
}



1;
