###########################################################################
# <LICENSEINFO>
# Copyright (C) Daniel Guermeur - 1999-2001
# Copyright (C) Metadot Corporation - 2001-2004
# 
# This file is part of Metadot Portal Server software $Name: Release_6_1_6 $
# 
# Metadot Portal Server software is free software; you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
# 
# Metadot Portal Server software is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# Metadot Portal Server; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# For more information, please contact info@metadot.com.
#</LICENSEINFO>
############################################################################

# ObjectAccessor is an persistence layer that interfaces to DBAccess.
# Rather than directly calling DBAccess, you should make equivalent
# calls to ObjectAccessor. In this way, you can create objects that are
# independent of the database implementation.
#
# The current implementation does not yet fully free you from the details
# of the database implementation. It is a work in progress.

# IMPORTANT: The objects being accessed must implement the Restorable interface,
#            and must know how to use the Memento interface.
#            ObjectAccessor differs from the classic Memento pattern's Caretaker
#            interface, in that it has a transparent view of the Memento objects.
#


package Metadot::ObjectAccess::ObjectAccessor;
use strict;
use DBAccess;
use Metadot qw($PARAMS);
use Metadot::ObjectAccess::Memento;
use Metadot::ObjectAccess::Restorable;
use Metadot::ObjectAccess::DistributedCache;
use Metadot::SystemMessage;
use Metadot::SystemMessageManager;
use MetadotConfig;
use Metadotclass;
use Digest::MD5;
use Data::Dumper;


sub get_cache {
    my $self = shift;
    my $result;
    if ($PARAMS->{accelerator_enabled}) {
        my $server_name = MetadotConfig->get_server_name();
        $result = $self->{cache}->{$server_name};
        unless ($result) {
            $result = Metadot::ObjectAccess::DistributedCache->new(
                properties => {
                   owner => $self,
                   signature => $self->get_signature(),
                },
            );
            $self->{cache}->{$server_name} = $result;
            Metadotclass->debug_msg(2, "ObjectAccessor->get_cache for server $server_name CREATES $result");
        }
        Metadotclass->debug_msg(5, "ObjectAccessor->get_cache for server $server_name returns $result");
    }
    return $result;
}

sub get_from_cache {
    my ($self, %args) = @_;
    defined(my $key = $args{key}) || die("Must specify the key for the cache entry");
    my $result;

    my $cache = $self->get_cache();
    if ($cache) {
        $result = $self->get_cache()->get($key);
    }
    return $result;
}

sub put_in_cache {
    my ($self, %args) = @_;
    defined(my $object = $args{object}) || die("Must specify the object to store in cache");
    defined(my $key = $args{key}) || die("Must specify the key for the cache entry");

    my $cache = $self->get_cache();
    if ($cache) {
        $cache->put($key, $object);
    }

    Metadotclass->debug_msg(5, "ObjectAccessor\->put_in_cache successfully stored a cached object $object with key=$key");
}

sub remove_from_cache {
    my ($self, %args) = @_;
    defined(my $key = $args{key}) || die("Must specify the key for the cache entry");
    
    my $cache = $self->get_cache();
    if ($cache) {
        $cache->delete($key);
    }
}

sub notify_caches {
    my $self = shift;
    defined (my $command = shift) || die("Must specify key of object that has changed");
    # commands supported: FLUSH
    #                     DROP [key]
    my $message = Metadot::SystemMessage->new(
        properties=>{
            sender  =>$self->get_signature(),
            content =>$command,
        },
    );
    Metadot::SystemMessageManager->send_message($message);
}


# The constructor arguments:
# - properties are the object's internal properties
# - tables are the database tables used to store the object
# - key_property is the single property that is used to
#   uniquely identify an object in the database
# - joins specify the joins between the various tables
# - field map allows you to specify which database field
#   maps to each object property. Entries are required if
#   the database field is not the exact same name as the
#   object property

sub new {
   my $proto = shift;
   my $class = ref($proto) || $proto;
   my $self = {
      owner         => undef,
      properties    => [],
      tables        => [],
      key_property  => undef,
      joins         => [],
      field_map     => {},
      owner         => undef,
   };

   bless ($self, $class);

   my %args = @_;
   if (%args) {
      $self->initialize(%args);
   }

   $self->generate_signature();
   return $self;
}

# initializes the ObjectAccessor instance based on constructor's arguments
sub initialize {
   my $self = shift;
   my %args = @_;

   # array properties
   foreach my $property ('properties','tables','joins') {
      if (defined($args{$property})) {
         $self->{$property} = $self->_force_to_array_ref($args{$property});
      }
   }

   # hashref properties
   foreach my $property ('field_map') {
      if (defined($args{$property})) {
         $self->{$property} = $self->_force_to_hash_ref($args{$property});
      }
   }

   # scalar properties
   foreach my $property ('key_property', 'owner', 'signature') {
       if (defined($args{$property})) {
          $self->{$property} = $args{$property};
       }
   }

   $self->{cache} = {};
}


sub generate_signature {
    my $self = shift;
    
    my $owner_class = ref($self->get_owner()) || $self->get_owner() || 'ANONYMOUS';
    my $class = ref($self) || $self;
    my @tables = $self->get_tables();
    my $result = "$owner_class:$class:TABLES:" . join(':', @tables);
    $result = Digest::MD5::md5_base64($result);

    $self->set_signature($result);
    return $result;
}

# loads an object's properties from the database. The object must
# be uniquely identified by a key value.
#
# This method must know the class to instantiate! It tries to determine
# this based on the following:
#
# - If the arg 'class_field' is provided, it tries to find that field in the
#   resulting data
# - If the arg 'class' is provided, it uses that class
# - Otherwise, it uses the class of the ObjectAccessor instance's owner

sub retrieve {
   my ($self, %args) = @_;
   my $key = $args{key};
   my $result;
   
   defined($key) or die("Must provide a key value");
   my $class = $args{class} || $self->get_owner();
   
   Metadotclass->debug_msg(4, "ObjectAccessor->retrieve with key=$key, class=$class");

   # try to pull from the temp-cache
   $result = $self->get_from_cache(key=>$key);
   Metadotclass->debug_msg(4, "ObjectAccessor->retrieve checks in cache, gets:$result");
   #print STDERR "ObjectAccessor->retrieve checks in cache, gets:$result\n";

   if ($result) {
       #print STDERR "ObjectAccessor->retrieve successfully retrieved with class=$class; key=$key\n";
   } else {
       my @fields = ();
       my @properties = $self->get_properties();

       # construct the sql statement
       foreach my $property (@properties) {
          push @fields, $self->get_field_for_property($property);
       }

       my $select_fields = join(',', @fields);
       my $select_from   = join(',', $self->get_tables());
       my @wheres = ($self->get_key_clause($key));
       if ($self->get_join_clause()) {
          push @wheres, $self->get_join_clause();
       }
       my $where_clause  = join(' AND ', @wheres);

       #print STDERR "ObjectAccessor->retrieve using: $select_fields\n$select_from\n$where_clause\n\n";

       my @values = DBAccess->sqlSelect( $select_fields, $select_from, $where_clause );

       my $memento = Metadot::ObjectAccess::Memento->new();
       # populate the memento from the results
       foreach my $property (@properties) {
          $memento->set_state($property, shift(@values));
       }

       $result = $self->_restore_instance(%args, memento=>$memento);
       $self->put_in_cache(object=>$result, key=>$key);
   }
   
   return $result;
}


# performs the actual job of restoring an instance based on a memento,
# and class args
#
# - If the arg 'class_field' is provided, it tries to find that field in the
#   memento
# - If the arg 'class' is provided, it uses that class
# - Otherwise, it uses the class of the ObjectAccessor instance's owner

sub _restore_instance {
    my $self = shift;
    my %args = @_;

    my $memento = $args{memento};
    my $class;
    my $owner = $self->get_owner();
    
    # first see if the class information can be detected via the memento. This is the
    # most reliable.
    if ($owner && $owner->can('determine_class_from_memento')) {
        $class = $owner->determine_class_from_memento($memento);
    }
     
    # otherwise, try to discern the class from the result. It looks for the 'class_field' property
    if (!$class && defined $args{class_field}) {
        $class = $memento->get_state($args{class_field});
    }

    # otherwise, check the 'class' property
    if (!$class && defined $args{class}) {
        $class = $args{class};
    }

    # finally, attempt to use the owner's class
    $class ||= ref($owner) || $owner;

    ($class) || die("Could not resolve the class to retrieve");
    Metadotclass->debug_msg(4, "ObjectAccessor will instantiate a new $class");

    my $result = $class->new();
    $result->restore_state($memento);

    return $result;
}


# retrieves one or more objects from the database, based on an SQL select query
sub retrieve_where {
    my ($self, %args) = @_;
    my $where = $args{where};
    my $class = $args{class} || $self->get_owner();

#print STDERR "ObjectAccessor->retrieve with class=$class and where=\n$where\n";
    
   my @results = ();
   my @fields = ();
   my @properties = ($self->get_properties());

   # construct the sql statement
   foreach my $property (@properties) {
      push @fields, $self->get_field_for_property($property);   
   }
      
   my $sth = $self->_retrieve_fields_where( where=>$where, fields=>\@fields);
   while (my @values = $sth->fetchrow()) {
       my $result;
       my $memento = Metadot::ObjectAccess::Memento->new();
       
       # populate the memento from the results
       foreach my $property (@properties) {
          $memento->set_state($property, shift(@values));
       }
       
       $result = $self->_restore_instance(%args, memento=>$memento);
       
       my $key_property    = $self->get_key_property();
       my $key_value       = $memento->get_state($key_property);
       
       $self->put_in_cache(object=>$result, key=>$key_value);
   }
   $sth->finish();

   if (wantarray) {
       return @results;
   } else {
       return \@results;
   }
}


sub update_where {
    my ($self, %args) = @_;
    my $where = $args{where};
    my $class = $args{class};
    my $updates = $args{updates};

    my $sth = $self->_retrieve_fields_where( where=>$where, fields=>[$self->get_key_field()]);
    while (my ($key) = $sth->fetchrow()) {
        my $object = $self->retrieve( %args, key=>$key );
        foreach my $key (keys %$updates) {
            #print STDERR "Updating field $key from '$object->{$key}' to '$updates->{$key}'\n";
            $object->{$key} = $updates->{$key};
        }
        $self->update( object=>$object );
    }
    $sth->finish();
}


# WARNING! THIS METHOD CURRENTLY SUPPORTS DELETING FROM ONE TABLE ONLY
sub delete_where {
    my ($self, %args) = @_;
    my $where = $args{where};
    my $updates = $args{updates};
    my $table   = $self->get_tables()->[0];

    my $sth = $self->_retrieve_fields_where( where=>$where, fields=>[$self->get_key_field()]);
    while (my ($key) = $sth->fetchrow()) {
        $self->remove_from_cache(key=>$key);
        $self->notify_caches("DROP $key");
    }
    $sth->finish();
    DBAccess->sqlDelete( $table, $where );
}


# returns an open statement handle
sub _retrieve_fields_where {
    my ($self, %args) = @_;
    defined(my $where = $args{where}) or die("Must supply a where clause");
    my $fields = $args{fields} || [$self->get_key_field()];

    my $select_fields = join(',', @$fields);
    my $select_from   = join(',', $self->get_tables());
    my @wheres = ($where);
    if ($self->get_join_clause()) {
       push @wheres, $self->get_join_clause();
    }
    my $where_clause  = join(' AND ', @wheres);

    #print STDERR "ObjectAccessor->retrieve using: $select_fields\n$select_from\n$where_clause\n\n";

    my $sth = DBAccess->sqlSelectMany( $select_fields, $select_from, $where_clause );
    return $sth;
}

# WARNING! THIS METHOD CURRENTLY SUPPORTS INSERTING INTO ONE TABLE ONLY
sub create {
    my ($self, %args) = @_;
    my $object          = $args{object} || die("Must supply the object to create");
    $object->isa('Metadot::ObjectAccess::Restorable') || die("Object must implement the Metadot::ObjectAccess::Restorable interface");
    
    my $table           = $self->get_tables()->[0];
    my $memento = $object->extract_state();

    my %values = ();
    foreach my $property ($self->get_properties()) {
        my $value = $memento->get_state($property);
        $values{$self->get_field_for_property($property)} = $value;
    }
    my $key_value = DBAccess->sqlInsert($table, %values);
    
    $self->put_in_cache(object=>$object, key=>$key_value);   
    return $key_value;
}


# WARNING! THIS METHOD CURRENTLY SUPPORTS UPDATING ONE TABLE ONLY
sub update {
    my ($self, %args) = @_;
    my $object = $args{object} || die("Must supply the object to update");
    $object->isa('Metadot::ObjectAccess::Restorable') || die("Object must implement the Metadot::ObjectAccess::Restorable interface");
    
    my $memento         = $object->extract_state();
    my $table           = $self->get_tables()->[0];
    my $key_property    = $self->get_key_property();
    my $key_value       = $memento->get_state($key_property);
    my $key_field       = $self->get_key_field();
    #print STDERR "ObjectAccessor->update  key property='$key_property'; key value='$key_value'; memento=\n",Dumper $memento; 

    my %values = ();
    foreach my $property ($self->get_properties()) {
        unless ($property eq $key_property) {
            my $value = $memento->get_state($property);
            $values{$self->get_field_for_property($property)} = $value;
        }
    }
    DBAccess->sqlUpdate($table, "$key_field=$key_value", %values);    
    $self->notify_caches("DROP $key_value");
    $self->put_in_cache(object=>$object, key=>$key_value);   
}


# WARNING! THIS METHOD CURRENTLY SUPPORTS UPDATING ONE TABLE ONLY
sub delete {
    my ($self, %args) = @_;
    defined(my $key_value = $args{key}) or die("Must provide a key value");
    my $object = $args{object};
    my $class = $args{class} || ref($object);
    
    my $table           = $self->get_tables()->[0];
    my $key_property    = $self->get_key_property();
    my $key_field       = $self->get_key_field();
    #print STDERR "ObjectAccessor->delete  key property='$key_property'; key value='$key_value';\n";

    DBAccess->sqlDelete($table, "$key_field=$key_value");    
    $self->notify_caches("DROP $key_value");
    $self->remove_from_cache(key=>$key_value);   
}


sub get_field_for_property {
   my ($self, $property) = @_;
   my $result;
   if (defined($self->{field_map}->{$property})) {
      $result = $self->{field_map}->{$property};
   } else {
      $result = $property;
   }
   return $result;
}

sub get_key_clause {
   my $self = shift;
   defined(my $key = shift) || die("Must provide the key value");
   my $result;

   my @wheres = ();
   foreach my $key_field ($self->get_key_field()) {
      push @wheres, "$key_field = $key";
   }

   $result = join(' AND ', @wheres);
   return $result;
}

sub get_join_clause {
   my $self = shift;
   my $result;

   my @wheres = ();
   foreach my $join ($self->get_joins()) {
      push @wheres, $join;
   }
   
   $result = join(' AND ', @wheres);
   return $result;
}

# accessors & mutators
sub get_properties {
   my $self = shift;
   my $result = $self->{properties};
   if (wantarray) {
       return $self->_force_to_array($result);
   } else {
       return $result;
   }
}

sub get_tables {
   my $self = shift;
   my $result = $self->{tables};
   if (wantarray) {
       return $self->_force_to_array($result);
   } else {
       return $result;
   }
}

# returns the object property that is the key
sub get_key_property {
   my $self = shift;
   return $self->{key_property};
}

# returns the database field that is the key
sub get_key_field {
    my $self = shift;
    my $result = $self->get_field_for_property($self->get_key_property());
    return $result;
}

sub get_owner {
    my $self = shift;
    return $self->{owner};
}

sub get_joins {
   my $self = shift;
   my $result = $self->{joins};
   if (wantarray) {
       return $self->_force_to_array($result);
   } else {
       return $result;
   }
}

sub set_properties {
   my $self = shift;
   my @values = @_;
   $self->{properties} = $self->_force_to_array_ref(@values);
}

sub set_tables {
   my $self = shift;
   my @values = @_;
   $self->{tables} = $self->_force_to_array_ref(@values);
}

sub set_key_property {
   my $self = shift;
   my $value = shift;
   $self->{key_property} = $value;
}

sub set_joins {
   my $self = shift;
   my @values = @_;
   $self->{joins} = $self->_force_to_array_ref(@values);
}


# private methods
sub _force_to_array {
   my $class = shift;
   my @input = @_;
   my @result;

   if (@input && scalar(@input) == 1 && ref($input[0]) eq 'ARRAY') {
      @result = @{ $input[0] };
   } else {
      @result = @input;
   }
   return @result;
}

sub _force_to_array_ref {
   my $class = shift;
   my @input = @_;
   my $result;
   
   if (@input && scalar(@input) == 1 && ref($input[0]) eq 'ARRAY') {
      $result = $input[0];
   } else {
      $result = [ @input ];
   }
   return $result;
}

sub _force_to_hash {
   my $class = shift;
   my @input = @_;
   my %result;

   if (@input && scalar(@input) == 1 && ref($input[0]) eq 'HASH') {
      %result = %{ $input[0] };
   } else {
      %result = @input;
   }
   return %result;
}

sub _force_to_hash_ref {
   my $class = shift;
   my @input = @_;
   my $result;
   
   if (@input && scalar(@input) == 1 && ref($input[0]) eq 'HASH') {
      $result = $input[0];
   } else {
      $result = { @input };
   }
   return $result;
}


sub get_signature {
    my $self = shift;
    return $self->{signature};
}

sub set_signature {
    my $self = shift;
    $self->{signature} = shift;
}

1;
