package Metadot::SkinManager;
use strict;

use Archive::Zip;
use XML::Simple;

use Metadot qw($PARAMS $DISPLAY %REQUEST_VARS);
use MetadotConfig;
use Metadot::Skin;
use Metadot::UploadsManager;
use Metadot::MetadotController;
use Metadotclass;

use constant _DEFAULT_SKIN_ID_    => 'Metadot6';
use constant _BASE_SKIN_ID_       => 'base_skin';
use constant _SKIN_DESCRIPTOR_  => 'skin.xml';

use constant _IMAGES_         => 'images';
use constant _STYLESHEETS_    => 'stylesheets';
use constant _TEMPLATES_      => 'templates';
use constant _SCRIPTS_        => 'scripts';


use constant _FILE_TYPES_ => [_IMAGES_, _STYLESHEETS_, _TEMPLATES_, _SCRIPTS_];


sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = {};

    return bless($self, $class);
}

# returns the file system root of all skin templates
sub get_skin_template_dir {
    my $class = shift;
    return MetadotConfig->get_skin_template_dir();
}


# scans the skins directory for available skins
sub get_available_skins {
    my $class = shift;
    my @files = $class->_get_directory_files($class->get_skin_template_dir());
    my @skins = ();
    foreach my $entry (@files) {
        #print STDERR "Looking for skin in '$entry'\n";
        my $skin = $class->get_skin($entry);
        #print STDERR "Got skin $skin\n";
        if ($skin) {
            push @skins, $skin;
        }
    }
    closedir DIRECTORY;
    
    if (wantarray) {
        return @skins;
    } else {
        return \@skins;
    }
}


# returns a single skin, based on an id
my %SKIN_CACHE;
sub get_skin {
    my $class = shift;
    my $skin_id = shift || die("Must specify skin id");
    
    my $skin;
    my $path = $class->get_skin_path($skin_id);
    my $descriptor_file = $path.'/'._SKIN_DESCRIPTOR_;

    if (-d "$path" && -e $descriptor_file) {
        my $file_date = (stat $descriptor_file)[9];
        Metadotclass->debug_msg(5, "Descriptor file '$descriptor_file' has date '$file_date'");
        # check the skin cache
        if (exists $SKIN_CACHE{$path}) {
            # check the file date
            if ($SKIN_CACHE{$path}{date} >= $file_date) {
                $skin = $SKIN_CACHE{$path}{skin};
                Metadotclass->debug_msg(5, "Retrieved skin $skin from cache '$path'");
            }
        }
        
        # no skin? read from the file system
        unless ($skin) {
            $skin = Metadot::Skin->new();
            $skin->set_id($skin_id);
            $class->apply_descriptor($skin);

            $SKIN_CACHE{$path}{skin} = $skin;
            $SKIN_CACHE{$path}{date} = $file_date;
            Metadotclass->debug_msg(5, "Added skin $skin to cache '$path' with date '$file_date'");
        }
    }

    return $skin;
}

sub drop_skin_cache {
    Metadotclass->debug_msg(5, "Dropping skin cache");
    %SKIN_CACHE = ();
    $REQUEST_VARS{skin_chains} = undef;
}


# returns a single skin, based on name. If there are multiple matches, only
# the first one is returned
sub get_skin_by_name {
    my $class = shift;
    my $skin_name = shift;
    my $result;

    if (defined $skin_name && $skin_name ne '') {
        my $available_skins = $class->get_available_skins();
        
        # check full names first
        foreach my $skin (@$available_skins) {
            if ($skin->get_name() eq $skin_name) {
                $result = $skin;
                last;
            }
        }
        
        # check short names if still not found
        unless ($result) {
            foreach my $skin (@$available_skins) {
                if ($skin->get_short_name() eq $skin_name) {
                    $result = $skin;
                    last;
                }
            }
        }
    }

    my $result_name;
    if ($result) {
        $result_name = $result->get_name();
    }
    Metadotclass->debug_msg(5, "get_skin_by_name($skin_name) returns $result ($result_name)");
    return $result;
}

# gets the id of the currently configured skin
sub get_current_skin_id {
    my $class = shift;
    my $result;
    if ($PARAMS->{skin_selection} && !Metadot::MetadotController->force_default_skin()) {
        $result = $PARAMS->{skin_selection};
    } else {
        $result = _DEFAULT_SKIN_ID_;
    }
    
    return $result;
}


# gets the currently configured skin
sub get_current_skin {
    my $class = shift;

    my $result = $class->get_skin($class->get_current_skin_id())
                || $class->get_skin(_DEFAULT_SKIN_ID_);

    return $result;
}


# creates a new skin.
# one will be computed
sub add_skin {
    my $class = shift;
    my $skin = shift;
    (defined($skin) && ref($skin) && $skin->isa('Metadot::Skin')) || die("Must pass a Metadot::Skin object");
    #print STDERR "in add_skin\n";
    
    my $skin_id = $class->_generate_id($skin->get_short_name() || $skin->get_name());
    $skin->set_id($skin_id);
    #print STDERR "generated skin_id $skin_id\n";

    my $template_dir = $class->get_skin_path($skin_id);
    #print STDERR "will create skin in $template_dir\n";

    if (mkdir($template_dir, 0755)) {
        mkdir("$template_dir/"._IMAGES_,        0755);
        mkdir("$template_dir/"._STYLESHEETS_,   0755);
        mkdir("$template_dir/"._TEMPLATES_,     0740);
        mkdir("$template_dir/"._SCRIPTS_,       0755);
        print STDERR "Finished writing directory structure\n";
    
        $class->write_descriptor($skin);
    } else {
        $DISPLAY->print_click_back('Error', "There was a problem creating the files for this skin in $template_dir. Please contact your system administrator");
        exit(0);
    }
    
    
    return $skin_id;
}

# writes the skin's descriptor
sub write_descriptor {
    my $class = shift;
    my $skin = shift;
    (defined($skin) && ref($skin) && $skin->isa('Metadot::Skin')) || die("Must pass a Metadot::Skin object");
    
    my $descriptor = $class->generate_descriptor($skin);
    my $template_dir = $class->get_skin_path($skin->get_id());

    if (open DESCRIPTOR_FILE, ">$template_dir/"._SKIN_DESCRIPTOR_) {
        print DESCRIPTOR_FILE $descriptor;
        close DESCRIPTOR_FILE;

        $class->drop_skin_cache();
    } else {
        $DISPLAY->print_click_back('Error', "There was a problem updating the skin in $template_dir. Please contact your system administrator");
        exit(0);
    }
}


# deletes the named skin
sub delete_skin {
    my $class = shift;
    my $skin_id = shift || die("Must specify skin id");
    ($skin_id ne _DEFAULT_SKIN_ID_ && $skin_id ne _BASE_SKIN_ID_) || die("Cannot delete the default skin");

    # check if read-only
    !$class->is_read_only_skin($skin_id) || die("Cannot delete this read-only skin");
    
    my $template_dir = Metadot::SkinManager->get_skin_template_dir();
    my $full_dir = $class->get_skin_path($skin_id);
    
    my $result = $class->_delete_directory($full_dir);
    
    $class->drop_skin_cache();

    return $result;
}


# recursively deletes the named directory
sub _delete_directory {
    my $class = shift;
    my $dir_name = shift || die("Must specify the directory name");
    my $result = 1;

    # get the files and subdirectories
    opendir DIRECTORY, $dir_name;
    my @subdirs = ();
    foreach my $entry (grep !/^\.\.?$/, readdir DIRECTORY) {
        #print STDERR "evaluating entry: $entry\n";
        if (-d "$dir_name/$entry") {
            push @subdirs, $entry;
        } else {
            (unlink "$dir_name/$entry");
        }
    }
    closedir DIRECTORY;

    # delete the subdirectories
    foreach my $entry (@subdirs) {
        $result &&= $class->_delete_directory($dir_name.'/'.$entry);
    }

    $result &&= (rmdir $dir_name);
    return $result;
}


# searches for and parses an XML description of the skin, returning a hash structure
sub apply_descriptor {
    my $class = shift;
    my $skin = shift;
    (defined($skin) && ref($skin) && $skin->isa('Metadot::Skin')) || die("Must pass a Metadot::Skin object");
        
    my $skin_id = $skin->get_id() || die("This skin does not have an id");
    
    my $template_dir = $class->get_skin_path($skin_id);
    my $filename = "$template_dir/"._SKIN_DESCRIPTOR_;
    my $result;

    #print STDERR "Checking file '$filename for skin attributes\n";

    # check for file
    if (-e $filename) {
        eval {
            $result = XMLin($filename);
        };
    }
    $result ||= {};

    $skin->set_properties($result);

    use Data::Dumper;
    #print STDERR "Skin attributes are:\n",Dumper $skin;

    return $skin;
}

sub generate_descriptor {
    my $class = shift;
    my $skin = shift;
    (defined($skin) && ref($skin) && $skin->isa('Metadot::Skin')) || die("Must pass a Metadot::Skin object");
        
    my $skin_id = $skin->get_id() || die("This skin does not have an id");
    
    my $template_dir = $class->get_skin_path($skin_id);
    my $filename = "$template_dir/"._SKIN_DESCRIPTOR_;
    my $attributes = {
        name        => $skin->get_name(),
        short_name  => $skin->get_short_name(),
        author      => $skin->get_author(),
        website     => $skin->get_website(),
        revision    => $skin->get_revision(),
        description => $skin->get_description(),
        license     => $skin->get_license(),
        based_on    => $skin->get_based_on(),
    };

    my $result = XMLout($attributes, RootName=>'skin');
    use Data::Dumper;
    #print STDERR "generated descriptor:\n\n$result\n\n from\n",Dumper($skin);

    return $result;
}

# converts the named skin to a zip file. Returns the name of the file that was created.
sub export_zip {
    my $class = shift;
    my $skin_id = shift || die("Must specify the skin id");
    my $file = shift || die("Must specify the desired file name");
    my $result;

    #print STDERR "Exporting skin '$skin_id' as file '$file'\n";

    my $template_dir = $class->get_skin_template_dir();
    my $skin_dir = $class->get_skin_path($skin_id);
    if (-e $skin_dir ) {
        my $read_file = $template_dir . '/' . $file;

        # try to create a .zip file
        #print STDERR "Creating zip $read_file, with directory $skin_dir\n";
        my $zip = Archive::Zip->new();
        $zip->addTree($skin_dir, $skin_id);
        $zip->writeToFileNamed($read_file);
        $class->_collect_garbage_skins();
        $result = $read_file;
    } else {
        $result = undef;
    }
    
    #print STDERR "Export completes with result '$result'\n";
    
    return $result;
}


# creates a skin from a zip file
sub import_zip {
    my $class = shift;
    my $skin_id = shift || die("Must supply a skin id");
    my $write_file = shift || die("Must supply a ZIP file");

    #print STDERR "import_zip begins with skin '$skin_id', file '$write_file'\n";
    
    ($skin_id ne _DEFAULT_SKIN_ID_ && $skin_id ne _BASE_SKIN_ID_) || return("Cannot overwrite the default skin");
    
    # check if read-only
    !$class->is_read_only_skin($skin_id) || return("Cannot delete overwrite read-only skin");
    
    my $template_dir = $class->get_skin_template_dir();
    my $skin_dir = $class->get_skin_path($skin_id);

    #print STDERR "target skin_dir is '$skin_dir'\n";

    my $zip = Archive::Zip->new($write_file);

    #print STDERR "I have zip object '$zip'\n";

    my $result = $zip->extractTree($skin_id, $skin_dir);
    #print STDERR "import_zip returns with result '$result'\n";
        
    $class->_collect_garbage_skins();

    $class->drop_skin_cache();

    return $result;
}


# deletes all zipped skin files that are older than 24 hours
sub _collect_garbage_skins {
    my $class = shift;

    my $path = $class->get_skin_template_dir();
    opendir SCRATCHDIR, $path;
    
    # the grep filters out . and ..
    my @file_list = grep !/^\.\.?$/, readdir SCRATCHDIR;

    closedir SCRATCHDIR;

    foreach my $file_base (@file_list) {
        if ($file_base =~ m/\.zip%/) {
            my $file_name = $path.'/'.$file_base;
            $^T = time; # reset script start time to allow easy file age comparison
            my $file_age = (-M $file_name);
            if ($file_age > 1) {
                unlink $file_name;
            }
        }
    }
}

# generates a mostly unique id for a skin. Provide a skin name as the base (optional).
sub _generate_id {
    my $class = shift;
    my $name = shift;
    my $id_base = $name || '';

    # first transform all whitespace and non-word characters
    $id_base =~ s/\s/_/g;
    $id_base =~ s/\W//g;

    # make sure there is some value by the end of this
    $id_base ||= 'skin';

    #print STDERR "generating id against '$name'; got base '$id_base'\n";
    # see if this id is in use, or keep trying to find an available id
    my $id = $id_base;
    my $sequence = 1;
    while (-e $class->get_skin_path($id)) {
        $id = $id_base.'_'.$sequence++;
    }

    return $id;
}


# returns the path for a given skin id. If no skin id is provided, the current will be used
sub get_skin_path {
    my $class = shift;
    my $skin_id = shift || $class->get_current_skin_id();

    my $result = $class->get_skin_template_dir().'/'.$skin_id;
    return $result;
}

# returns the base url for a given skin id. If no skin id is provided, the current will be used
sub get_skin_base_url {
    my $class = shift;
    my $skin_id = shift || $class->get_current_skin_id();

    my $result = "/skins/$skin_id";
    return $result;
}


# returns the path for a given skin file.
sub get_file_path {
    my $class = shift;
    my $args = shift || {};
    
    my $skin_id = $args->{skin_id} || $class->get_current_skin_id();
    my $name    = $args->{name};
    my $type    = $args->{type}    || $class->_determine_file_type($name);

    my $result = $class->get_skin_path($skin_id);
    if ($type) {
        $result .= '/'.$type;
        if ($name) {
            $result .= '/'.$name;
        }
    }

    return $result;
}

# returns boolean indicating whether the selected skin is the default
sub is_default_skin {
    my $class = shift;
    my $skin_id = shift || $class->get_current_skin_id();

    my $result = 0;
    if (defined $skin_id) {
        $result = ($skin_id eq _DEFAULT_SKIN_ID_);
    }

    return $result;
}

# returns boolean indicating whether the selected skin is read-only
sub is_read_only_skin {
    my $class = shift;
    my $skin_id = shift || $class->get_current_skin_id();

    my $skin = $class->get_skin($skin_id);
    return ($skin && $skin->is_read_only());
}


# returns a full INCLUDE path for a template
sub get_template_include_path {
    my $class = shift;
    my $skin_id = shift || $class->get_current_skin_id();
    
    my @INCLUDE_PATH = ();
    my $subpath = '/'._TEMPLATES_;
    
    
    # if the default skin is being forced, that's all
    if (Metadot::MetadotController->force_default_skin()) {
        @INCLUDE_PATH = $class->get_skin_path(_DEFAULT_SKIN_ID_).$subpath;
    } else {
        my $skin_chain = $class->get_skin_chain();
        foreach my $skin (@$skin_chain) {
            push @INCLUDE_PATH, $class->get_skin_path($skin->get_id()).$subpath;
        }
    }
    
    #print STDERR "get_template_include_path returns @INCLUDE_PATH\n";

    if (wantarray) {
        return @INCLUDE_PATH;
    } else {
        return \@INCLUDE_PATH;
    }
}

# returns the parent of this skin
sub get_parent_of_skin {
    my $class = shift;
    my $skin  = shift;
    
    my $result;
    
    # if I didn't get a skin, see if I can resolve it as a skin id
    if ($skin && !ref($skin)) {
        $skin = $class->get_skin($skin);
    }

    if ($skin) {
        # see if I am based on a skin
        my $based_on = $skin->get_based_on();
        if (defined $based_on && $based_on ne '') {
            $result = $class->get_skin_by_name($skin->get_based_on());
        }

        # if not based on, or returned nothing, get the default
        if (!$result && $skin->get_id() ne _BASE_SKIN_ID_) {
            $result = $class->get_skin(_BASE_SKIN_ID_);
        }
    
        my $result_name;
        if ($result) {
            $result_name = $result->get_name();
        }
        Metadotclass->debug_msg(5, "Get parent of skin (".$skin->get_name().") returns $result ($result_name)");
    }
    
    return $result;
}


# returns the chain of skin responsibility
sub get_skin_chain {
    my $class   = shift;
    my $args    = shift || {};
    my $visited = shift || {};
    my @result;

    my $skin    = $args->{skin};
    my $reverse = $args->{'reverse'};
    
    # check the request vars
    my $request_var_key = "$skin.$reverse";
    if (exists $REQUEST_VARS{skin_chains} && exists $REQUEST_VARS{skin_chains}{$request_var_key}) {
        @result = @{$REQUEST_VARS{skin_chains}{$request_var_key}};
    } else {
        # normal processing
        # see if I got an id instead
        if ($skin && !ref($skin)) {
            $skin = $class->get_skin($skin);
        }

        # no skin? get the current
        $skin ||= $class->get_current_skin();

        if ($skin) {
            # first get my parents
            if (my $parent_skin = $class->get_parent_of_skin($skin)) {
                my $parent_id = $parent_skin->get_id();
                if (!$visited->{$parent_id}) {
                    $visited->{$parent_id} = 1;
                    $args->{skin} = $parent_skin;
                    @result = $class->get_skin_chain($args, $visited);
                }
            }

            # then add myself
            if ($reverse) {
                push @result, $skin;
            } else {
                unshift @result, $skin;
            }
            #print STDERR "Added $skin to chain ($skin->{id})\n";
        }

        # update the request vars
        if (@result) {
            $REQUEST_VARS{skin_chains}{$request_var_key} = \@result;
        }
    }


    if (wantarray) {
        return @result;
    } else {
        return \@result;
    }
}


# returns a list of filenames within a given subdirectory and skin
sub get_filelist_for_type {
    my $class = shift;
    my $skin_id = shift || $class->get_current_skin_id();
    my $element_type = shift || die("Must specify the file type: ".@{&_FILE_TYPES_});

    my $directory = $class->get_skin_path($skin_id)."/".$element_type;
    my @files = ();
    foreach my $file ($class->_get_directory_files($directory)) {
        unless (-d "$directory/$file") {
            push @files, $file;
        }
    }

    if (wantarray) {
        return @files;
    } else {
        return \@files;
    }
}

# returns the URL for a file. Considers inheritance.
sub get_file_url {
    my $class = shift;
    my $args = shift || {};
    
    my $skin_id = $args->{skin_id} || $class->get_current_skin_id();
    my $name    = $args->{name};
    my $type    = $args->{type}    || $class->_determine_file_type($name);

    my $result = undef;
    my $skin_chain = $class->get_skin_chain({skin=>$skin_id});
    foreach my $skin (@$skin_chain) {
        if (-e $class->get_file_path({skin_id => $skin->get_id(), type=>$type, name=>$name})) {
            $result = $class->get_skin_base_url($skin->get_id())."/$type/$name";
            last;
        }
    }
    return $result;
}


# returns the contents for a file. Considers inheritance.
sub get_file_contents {
    my $class   = shift;
    my $args    = shift || {};
    
    my $skin_id = $args->{skin_id} || $class->get_current_skin_id();
    my $name    = $args->{name}    || die("Must specify arg name");
    my $type    = $args->{type}    || $class->_determine_file_type($name);

    my $result = undef;
    my $skin_chain = $class->get_skin_chain({skin=>$skin_id});
    foreach my $skin (@$skin_chain) {
        if ($class->file_exists({skin_id=>$skin->get_id(), type=>$type, name=>$name})) {
            open SKIN_FILE, $class->get_file_path({skin_id=>$skin->get_id(), type=>$type, name=>$name});
            my $buf = undef;
            while (read SKIN_FILE, $buf, 16384) {
                $result .= $buf;
            }
            close SKIN_FILE;
            last;
        }
    }

    return $result;
}


# determines if a file exists for a given skin
sub file_exists {
    my $class = shift;
    my $args = shift || {};
    
    my $skin_id = $args->{skin_id} || die("Must specify arg skin_id");
    my $type    = $args->{type}    || die("Must specify arg type");
    my $name    = $args->{name}    || die("Must specify arg name");

    return (-e $class->get_file_path($args));
}


# delete a single file from a skin
sub delete_file {
    my $class = shift;
    my $args = shift || {};
    
    my $skin_id = $args->{skin_id} || die("Must specify arg skin_id");
    my $type    = $args->{type}    || die("Must specify arg type");
    my $name    = $args->{name}    || die("Must specify arg name");

    !$class->is_read_only_skin($skin_id) || die("The skin '$skin_id' is read-only");

    unlink($class->get_file_path($args));
}


# save a single file to a skin
sub update_file {
    my $class = shift;
    my $args = shift || {};
    
    my $skin_id = $args->{skin_id} || die("Must specify arg skin_id");
    my $type    = $args->{type}    || die("Must specify arg type");
    my $name    = $args->{name}    || die("Must specify arg name");
    my $value   = $args->{value}   || '';

    !$class->is_read_only_skin($skin_id) || die("The skin '$skin_id' is read-only");
    
    my $filename = $class->get_file_path($args);

    #print STDERR "Will update '$filename' with content:\n$value\n----\n";
    
    open WRITE_FILE, '>'.$filename;
    print WRITE_FILE $value;
    close WRITE_FILE;

}

# upload a single file into the skin
sub upload_file {
    my $class = shift;
    my $args = shift || {};
    
    my $skin_id = $args->{skin_id} || die("Must specify arg skin_id");
    my $type    = $args->{type}    || die("Must specify arg type");
    my $name    = $args->{name}    || die("Must specify arg name");

    !$class->is_read_only_skin($skin_id) || die("The skin '$skin_id' is read-only");
    
    my $path = $class->get_file_path({skin_id=>$skin_id, type=>$type});
    # check that the path exists
    unless (-d $path) {
        mkdir($path, 0755);
    }

    my $result = Metadot::UploadsManager->upload_file(
        source_file => $name,
        write_dir   => $path,
        overwrite   => 1,
    );

    #print STDERR "SkinManager->upload_file gets result: $result\n";
    
    return $result;
}


sub _get_directory_files {
    my $class = shift;
    my $directory = shift || die("Must specify directory");

    opendir DIRECTORY, $directory;
    #my @files = (grep !/^\.\.?$/, readdir DIRECTORY);
    # trying new logic, which will also suppress .filename
    my @files = (grep !/^\./, readdir DIRECTORY);
    closedir DIRECTORY;
    
    if (wantarray) {
        return @files;
    } else {
        return \@files;
    }
}

# tries to determine the file type based on its filename
sub _determine_file_type {
    my $class = shift;
    my $file_name = shift;

    my $result;
    if (!$file_name) {
        $result = undef;
    } elsif ($file_name =~ /\.tt2$/) {
        $result = _TEMPLATES_;
    } elsif ($file_name =~ /\.css$/) {
        $result = _STYLESHEETS_;
    } elsif ($file_name =~ /\.js$/) {
        $result = _SCRIPTS_;
    } else {
        $result = _IMAGES_;
    }

    return $result;
}



1;
